
from germanium.selectors import *

# germanium instantiation
from .open_browser import open_browser
from .close_browser import close_browser
from .get_germanium import get_germanium
from .get_web_driver import get_web_driver

# germanium instance functions
from .click import click
from .double_click import double_click
from .go_to import go_to
from .hover import hover
from .right_click import right_click
from .S import S
from .type_keys import type_keys
from .js import js
from .get_attributes import get_attributes

# decorators
from .iframe import iframe