
from .CompositeLocator import CompositeLocator
from .CssLocator import CssLocator
from .DeferredLocator import DeferredLocator
from .SimpleLocator import SimpleLocator
from .StaticElementLocator import StaticElementLocator
from .XPathLocator import XPathLocator
from .JsLocator import JsLocator

from .PositionalFilterLocator import PositionalFilterLocator
